from PIL import Image

from . import _pycharls
from ._pycharls import (
    SpiffProfileId,
    SpiffCompressionType,
    SpiffColorSpace,
    SpiffResolutionUnits,
    SpiffHeader
)


_mode_table = {
    "1": (1, 8, SpiffColorSpace.BiLevelBlack),
    "L": (1, 8, SpiffColorSpace.Grayscale),
    "P": (1, 8, SpiffColorSpace.Grayscale),
    "RGB": (3, 8, SpiffColorSpace.Rgb),
    "RGBA": (4, 8, SpiffColorSpace.Rgb),
    "LAB": (3, 8, SpiffColorSpace.CieLab),
    "CMYK": (4, 8, SpiffColorSpace.Cmyk),
    "YCbCr": (3, 8, SpiffColorSpace.YCbCrItuBt601Video),
    "I;16": (1, 16, SpiffColorSpace.Grayscale),
    "RGB;16": (3, 16, SpiffColorSpace.Rgb),
}


def modes_to_str():
    return ", ".join(_mode_table.keys())


def _spiff_header(image: Image, spiff_header):
    component_count, bits_per_sample, default_color_space = _mode_table[image.mode]
    spiff = SpiffHeader()
    spiff.profile_id = spiff_header.get("profile_id", SpiffProfileId.NotSpecified)
    spiff.component_count = component_count
    spiff.height = image.height
    spiff.width = image.width
    spiff.color_space = spiff_header.get("color_space", default_color_space)
    spiff.bits_per_sample = max(bits_per_sample, 2)
    spiff.compression_type = SpiffCompressionType.JpegLs
    spiff.resolution_units = spiff_header.get("resolution_units", SpiffResolutionUnits.DotsPerInch)
    spiff.vertical_resolution = spiff_header.get("vertical_resolution", 96)
    spiff.horizontal_resolution = spiff_header.get("horizontal_resolution", 96)
    return spiff


def save(image: Image, fp, file_name):
    if image.mode not in _mode_table:
        raise SyntaxError(
            f"Image mode {image.mode} not supported. Convert to one of the supported modes: {modes_to_str()}.")

    component_count, bits_per_sample, _ = _mode_table[image.mode]
    frame_info = _pycharls.FrameInfo()
    frame_info.width = image.width
    frame_info.height = image.height
    frame_info.component_count = component_count
    frame_info.bits_per_sample = max(image.encoderinfo.get("bits_per_sample", bits_per_sample), 2)

    spiff_header = image.encoderinfo.get("spiff", {})
    header = None
    if spiff_header is not None:
        header = _spiff_header(image, spiff_header)
        header.bits_per_sample = frame_info.bits_per_sample

    image.encoderinfo["spiff"] = header
    if image.mode == "1":
        image.encoderinfo["maxval"] = 1
        # bitmap images are stored in a compressed representation if you use tobytes()
        data = bytes(image.getdata())
    else:
        data = image.tobytes()

    fp.write(_pycharls.encode(data, frame_info, **image.encoderinfo))



from PIL.ImageFile import PyDecoder

from . import _pycharls
from ._pycharls import SpiffColorSpace


class JplsImageDecoder(PyDecoder):
    def __init__(self, mode, *args):
        super().__init__(mode, *args)
        self._pulls_fd = True  # we decode the entire image in one shot
        self.color_space = args[0].get("color_space")

    def decode(self, buffer):
        buffer = self.fd.read()
        errcode = 0
        try:
            decoded = _pycharls.decode(buffer)
            raw_mode = None
            if self.color_space in [SpiffColorSpace.BiLevelWhite, SpiffColorSpace.BiLevelBlack]:
                raw_mode = "1;8"
            self.set_as_raw(bytes(decoded), rawmode=raw_mode)
        except RuntimeError:
            errcode = -2

        return -1, errcode



import os
from io import BytesIO

import pillow_jpls # noqa
from PIL import Image
import pytest
import numpy as np


@pytest.mark.profile
def test_roundtrip(data):
    src = Image.open(os.path.join(data, "artificial.ppm"))

    buffer = BytesIO()
    src.save(buffer, format="JPEG-LS")

    buffer.seek(0)
    dest = Image.open(buffer)

    src_data = np.array(src, dtype=np.int32)
    dest_data = np.array(dest, dtype=np.int32)

    assert np.linalg.norm(src_data.ravel() - dest_data.ravel(), ord=np.inf) == 0



from PIL.ImageFile import ImageFile

from . import _pycharls
from ._pycharls import SpiffColorSpace


def accept(header):
    magic_number = header[:4]
    spiff_header = b"\xff\xd8\xff\xe8"
    jpeg_header = b"\xff\xd8\xff\xf7"
    return magic_number == spiff_header or magic_number == jpeg_header


def _mode(num_components, bits_per_component):
    if bits_per_component > 8:
        bits_per_component = 16

    mode_table = {
        (1, 1): "1",
        (3, 8): "RGB",
        (4, 8): "RGBA",
        (1, 8): "L",
        (1, 16): "I;16"
    }
    return mode_table.get((num_components, bits_per_component))


def _metadata(header):
    meta = {
        "bits_per_sample": header.bits_per_sample,
        "component_count": header.component_count,
        "width": header.width,
        "height": header.height
    }
    if hasattr(header, "profile_id"):
        meta.update({
            "color_space": header.color_space,
            "profile_id": header.profile_id,
            "compression_type": header.compression_type,
            "resolution_units": header.resolution_units,
            "vertical_resolution": header.vertical_resolution,
            "horizontal_resolution": header.horizontal_resolution
        })
    return meta


class JplsImageFile(ImageFile):
    format = "JPEG-LS"
    format_description = "Lossless and near-lossless compression of continuous-tone still images"

    def _open(self):
        max_header_bytes = 128
        buffer = self.fp.read(max_header_bytes)
        header = _pycharls.read_header(buffer)
        mode = _mode(header.component_count, header.bits_per_sample)
        if mode is None:
            raise IOError(f"Mode not supported: components: {header.component_count}, bits: {header.bits_per_sample}")

        meta = _metadata(header)
        color_space = meta.get("color_space")
        if color_space in [SpiffColorSpace.BiLevelBlack, SpiffColorSpace.BiLevelWhite]:
            mode = "1"
        elif color_space == SpiffColorSpace.CieLab:
            mode = "LAB"
        elif color_space == SpiffColorSpace.Cmyk:
            mode = "CMYK"
        elif color_space in [SpiffColorSpace.YCbCrItuBt601Video,
                             SpiffColorSpace.YCbCrItuBt601Rgb,
                             SpiffColorSpace.YCbCrItuBt709Video]:
            mode = "YCbCr"

        self.info.update(_metadata(header))
        self._size = (header.width, header.height)
        self.mode = mode
        self.tile = [("jpeg_ls", (0, 0) + self.size, 0, (meta, ))]



from PIL import Image
from .jpls_image_file import JplsImageFile, accept
from .jpls_image_encoder import save
from .jpls_image_decoder import JplsImageDecoder

from ._pycharls import ( # noqa
    SpiffProfileId,
    SpiffResolutionUnits,
    SpiffColorSpace,
    SpiffCompressionType
)

Image.register_open(JplsImageFile.format, JplsImageFile, accept)
Image.register_extensions(JplsImageFile.format, [".jls", ".jpls"])
Image.register_save(JplsImageFile.format, save)
Image.register_decoder("jpeg_ls", JplsImageDecoder)
Image.register_mime(JplsImageFile.format, "image/jls")



from io import BytesIO
import os

from PIL import Image
import pytest
import numpy as np

import pillow_jpls # noqa # pylint: disable=unused-import

# Conformance tests from Annexe E of the `ITU Spec <https://www.itu.int/rec/T-REC-T.87/en>`_.
# Tests 7 and 8 are omitted as subsampling is not supported by CharLS (and is rarely used)


def _encode_to_array(file_name, mode=None, **kwargs):
    src = Image.open(file_name)
    if mode is not None:
        src = src.convert(mode)

    buffer = BytesIO()
    src.save(buffer, format="JPEG-LS", spiff=None, **kwargs)
    return np.frombuffer(buffer.getvalue(), dtype=np.uint8).astype(np.int32)


def _decode_to_array(file_name, mode=None):
    image = Image.open(file_name)
    if image.mode == "I":
        image.convert("I;16")
    elif mode is not None:
        image = image.convert(mode)

    return np.array(image, dtype=np.int32)


def _load_encoded_data(file_name):
    with open(file_name, "rb") as fp:
        return np.frombuffer(fp.read(), dtype=np.uint8).astype(dtype=np.int32)


def assert_array_equal(expected: np.ndarray, actual: np.ndarray, epsilon=0):
    assert np.linalg.norm(expected.ravel() - actual.ravel(), ord=np.inf) <= epsilon


def test_01_compress_lossless_no_interleave(data):
    encoded_data = _encode_to_array(os.path.join(data, "TEST8.png"), interleave="none")
    expected_data = _load_encoded_data(os.path.join(data, "T8C0E0.JLS"))

    assert_array_equal(expected_data, encoded_data)


def test_02_compress_lossless_line_interleave(data):
    encoded_data = _encode_to_array(os.path.join(data, "TEST8.png"), interleave="line")
    expected_data = _load_encoded_data(os.path.join(data, "T8C1E0.JLS"))

    assert_array_equal(expected_data, encoded_data)


def test_03_compress_lossless_sample_interleave(data):
    encoded_data = _encode_to_array(os.path.join(data, "TEST8.png"))
    expected_data = _load_encoded_data(os.path.join(data, "T8C2E0.JLS"))

    assert_array_equal(expected_data, encoded_data)


def test_04_compress_lossy_no_interleave(data):
    encoded_data = _encode_to_array(os.path.join(data, "TEST8.png"), near_lossless=3, interleave="none")
    expected_data = _load_encoded_data(os.path.join(data, "T8C0E3.JLS"))

    assert_array_equal(expected_data, encoded_data)


def test_05_compress_lossy_line_interleave(data):
    encoded_data = _encode_to_array(os.path.join(data, "TEST8.png"), near_lossless=3, interleave="line")
    expected_data = _load_encoded_data(os.path.join(data, "T8C1E3.JLS"))

    assert_array_equal(expected_data, encoded_data)


def test_06_compress_lossy_sample_interleave(data):
    encoded_data = _encode_to_array(os.path.join(data, "TEST8.png"), near_lossless=3)
    expected_data = _load_encoded_data(os.path.join(data, "T8C2E3.JLS"))

    assert_array_equal(expected_data, encoded_data)


def test_09_compress_lossless_custom_threshold(data):
    encoded_data = _encode_to_array(os.path.join(data, "TEST8BS2.png"), t1=9, t2=9, t3=9, reset=31)
    expected_data = _load_encoded_data(os.path.join(data, "T8NDE0.JLS"))

    assert_array_equal(expected_data, encoded_data)


def test_10_compress_lossy_custom_threshold(data):
    encoded_data = _encode_to_array(os.path.join(data, "TEST8BS2.png"), near_lossless=3, t1=9, t2=9, t3=9, reset=31)
    expected_data = _load_encoded_data(os.path.join(data, "T8NDE3.JLS"))

    assert_array_equal(expected_data, encoded_data)


def test_11_compress_lossless_16(data):
    encoded_data = _encode_to_array(os.path.join(data, "TEST16.png"), mode="I;16", bits_per_sample=12)
    expected_data = _load_encoded_data(os.path.join(data, "T16E0.JLS"))

    assert_array_equal(expected_data, encoded_data)


def test_12_compress_lossy_16(data):
    encoded_data = _encode_to_array(os.path.join(data, "TEST16.png"), mode="I;16", near_lossless=3, bits_per_sample=12)
    expected_data = _load_encoded_data(os.path.join(data, "T16E3.JLS"))

    assert_array_equal(expected_data, encoded_data)


@pytest.mark.parametrize("encoded, decoded, epsilon", [
    ("T8C0E0.JLS", "TEST8.png", 0),
    ("T8C1E0.JLS", "TEST8.png", 0),
    ("T8C2E0.JLS", "TEST8.png", 0),
    ("T8C0E3.JLS", "TEST8.png", 3),
    ("T8C1E3.JLS", "TEST8.png", 3),
    ("T8C2E3.JLS", "TEST8.png", 3),
    ("T8NDE0.JLS", "TEST8BS2.png", 0),
    ("T8NDE3.JLS", "TEST8BS2.png", 3),
    ("T16E0.JLS", "TEST16.png", 0),
    ("T16E3.JLS", "TEST16.png", 3)])
def test_decompress(data, encoded, decoded, epsilon):
    jls_decoded = _decode_to_array(os.path.join(data, encoded))
    raw_decoded = _decode_to_array(os.path.join(data, decoded))

    assert_array_equal(raw_decoded, jls_decoded, epsilon=epsilon)



from io import BytesIO

from PIL import Image
import numpy as np
import pytest

from pillow_jpls import (
    SpiffProfileId,
    SpiffResolutionUnits,
    SpiffColorSpace,
    SpiffCompressionType
)

_rng = np.random.default_rng()


def test_bilevel_image():
    width = 300
    height = 200
    data = _rng.integers(2, size=(height, width), dtype=np.uint8)
    buffer = BytesIO()
    Image.fromarray(data * 255, mode="L").convert("1").save(buffer, format="JPEG-LS")

    buffer.seek(0)
    decoded_data = np.array(Image.open(buffer)).astype(np.uint8)

    assert np.linalg.norm(decoded_data - data, ord=np.inf) == 0


@pytest.mark.parametrize("shape, bits, maxval, dtype, mode, colorspace",[
    ((42, 23), 8, 255, np.uint8, "L", SpiffColorSpace.Grayscale),
    ((18, 6), 8, 255, np.uint8, "P", SpiffColorSpace.Grayscale),
    ((8, 7, 3), 8, 255, np.uint8, "RGB", SpiffColorSpace.Rgb),
    ((365, 7, 4), 8, 255, np.uint8, "RGBA", SpiffColorSpace.Rgb),
    ((12, 47, 3), 8, 255, np.uint8, "LAB", SpiffColorSpace.CieLab),
    ((12, 507, 4), 8, 255, np.uint8, "CMYK", SpiffColorSpace.Cmyk),
    ((19, 87, 3), 8, 255, np.uint8, "YCbCr", SpiffColorSpace.YCbCrItuBt601Video),
    ((52, 13), 16, 65535, np.uint16, "I;16", SpiffColorSpace.Grayscale),
    ((3, 4), 12, 4095, np.uint16, "I;16", SpiffColorSpace.Grayscale)
])
def test_encode_with_spiff_header(shape, bits, maxval, dtype, mode, colorspace):
    width = shape[1]
    height = shape[0]
    component_count = 1 if len(shape) == 2 else shape[-1]
    data = _rng.integers(maxval + 1, size=shape, dtype=dtype)

    buffer = BytesIO()
    Image.fromarray(data, mode).save(buffer, format="JPEG-LS", bits_per_sample=bits)

    buffer.seek(0)
    decoded_image = Image.open(buffer)
    decoded_data = np.array(decoded_image, dtype=dtype)

    expected_header = {
        "profile_id": SpiffProfileId.NotSpecified,
        "component_count": component_count,
        "width": width,
        "height": height,
        "color_space": colorspace,
        "bits_per_sample": max(bits, 2),
        "compression_type": SpiffCompressionType.JpegLs,
        "resolution_units": SpiffResolutionUnits.DotsPerInch,
        "vertical_resolution": 96,
        "horizontal_resolution": 96
    }

    assert decoded_image.info == expected_header
    assert np.linalg.norm(data.ravel() - decoded_data.ravel(), ord=np.inf) == 0



import os

import pytest


_profile_keyword = "profile"


def pytest_addoption(parser):
    default_data_path = os.path.realpath(os.path.join(os.path.dirname(__file__), "data"))
    parser.addoption( f"--{_profile_keyword}", action="store_true", default=False, help="run profiling tests" )
    parser.addoption( "--data", action="store", default=default_data_path, help="data folder location")


def pytest_configure(config):
    config.addinivalue_line("markers", f"{_profile_keyword}: mark test as profiling test")


def pytest_collection_modifyitems(config, items):
    if config.getoption(f"--{_profile_keyword}"):
        return
    skip_profiling = pytest.mark.skip(reason="profiling test")
    for item in items:
        if _profile_keyword in item.keywords:
            item.add_marker(skip_profiling)


@pytest.fixture
def data(request):
    return os.path.realpath(request.config.getoption("--data"))



import sys
from os import path
from setuptools import find_packages
import re


_semver_regex = r"""(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)"""
_project_regex = r"^project\(.*$"

try:
    from skbuild import setup
except ImportError:
    print(
        "Please update pip, you need pip 10 or greater,\n"
        " or you need to install the PEP 518 requirements in pyproject.toml yourself",
        file=sys.stderr,
    )
    raise


def _current_directory():
    return path.abspath(path.dirname(__file__))


def _long_description():
    with open(path.join(_current_directory(), 'README.md'), encoding='utf-8') as fp:
        return fp.read()


def _version():
    with open(path.join(_current_directory(), 'CMakeLists.txt'), encoding='utf-8') as fp:
        text = fp.read()
        project_line = re.search(_project_regex, text, re.MULTILINE | re.IGNORECASE)
        semver = re.search(_semver_regex, project_line.group())
        return "{major}.{minor}.{patch}".format(**semver.groupdict())


setup(
    name="pillow-jpls",
    version=_version(),
    description="A JPEG-LS plugin for the Pillow imaging library",
    author="Andrew Marshall",
    author_email="andrew@algodynamic.com",
    url="https://github.com/planetmarshall/pillow-jpls",
    license="BSD-3-Clause",
    long_description=_long_description(),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    cmake_install_dir="pillow_jpls",
    install_requires=[
        "Pillow"
    ],
    python_requires=">=3.6"
)